/*
 *  Uzebox Video Poker R (Cosmetic Revamp)
 *  Copyright (C) 2013 Jim Croskell
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/****************************************
 *					Library Dependencies				*
 ****************************************/
#include <stdbool.h>
#include <avr/io.h>
#include <stdlib.h>
#include <avr/pgmspace.h>
#include <uzebox.h>

/****************************************
 *			  	 				Data								*
 ****************************************/
// Tiles
#include "data/tiles.inc"

// Font
#include "data/font6x8.inc"

// Sprites

// Sound
#include "data/music.inc"
#include "data/hi_score.inc"
#include "data/patches.inc"

/****************************************
 *			  	 Defines				*
 ****************************************/
#define HZ 60
#define GAME_ID 52

#define CREDITS_START 100
#define SCORE_START 0
#define BET_MIN 1
#define BET_MAX 5

#define DECK_SIZE 52
#define DECK_POS_X 32
#define DECK_POS_Y 17

#define HAND_SIZE 5

#define PAYOUT_TEXT_X 2
#define PAYOUT_TEXT_Y 13
#define PAYOUT_AMOUNT_X 21
#define PAYOUT_AMOUNT_Y 13

#define CREDIT_TEXT_X 3
#define CREDIT_TEXT_Y 25
#define CREDIT_VAL_X 12
#define CREDIT_VAL_Y 25

#define BET_TEXT_X 18
#define BET_TEXT_Y 25
#define BET_VAL_X 22
#define BET_VAL_Y 25

#define WIN_TEXT_X 28
#define WIN_TEXT_Y 25
#define WIN_VAL_X 36
#define WIN_VAL_Y 25

#define HELP_X_POS 24
#define HELP_Y_POS 5
#define HELP_X_SIZE 14
#define HELP_Y_SIZE 9

#define CARD_CURSOR_X_POS 5
#define CARD_CURSOR_Y_POS 22
//Card Suits
#define SPADES 0
#define HEARTS 1
#define DIAMONDS 2
#define CLUBS 3

// Card Values
#define ACE 0
#define TWO 1
#define THREE 2
#define FOUR 3
#define FIVE 4
#define SIX 5
#define SEVEN 6
#define EIGHT 7
#define NINE 8
#define TEN 9
#define JACK 10
#define QUEEN 11
#define KING 12

#define ROYAL_FLUSH 9
#define STRAIGHT_FLUSH 8
#define FOUR_OF_A_KIND 7
#define FULL_HOUSE 6
#define FLUSH 5
#define STRAIGHT 4
#define THREE_OF_A_KIND 3
#define TWO_PAIR 2
#define JACKS_OR_BETTER 1
#define LOSE 0

#define CARD_BACK 52
#define CARD_BLANK 53

#define MAX_PAYOUT 9

#ifndef CHAR_ZERO
	#define CHAR_ZERO 16 
#endif

// Utility macros
#define MIN(x, y) ((x) < (y) ? (x) : (y))
#define MAX(x, y) ((x) > (y) ? (x) : (y))
#define PRNG_NEXT() (prng = ((u8)((prng>>1) | ((prng^(prng>>2)^(prng>>3)^(prng>>4))<<7))))

/****************************************
 *			Type declarations			*
 ****************************************/
struct EepromBlockStruct ebs;

typedef enum
	{ 
		titleDraw, titleWait, playStart, playing, highScore, gameOver
	} GameState;

typedef struct
	{
		u8 x;
		u8 y;
	} Card_XY;

typedef struct
	{
		u8 card;
		u8 suit;
		u8 rank;
		bool held;
	} Hand;
	
/****************************************
 *			File-level variables		*
 ****************************************/
GameState gameState;
u8 FlashState = 1;
u8 SwapCardsState = 1;
u8 prng = 1;
u8 vram_buff[40];
u8 deck_pos = 0;
u8 deck[DECK_SIZE];

Hand hand[HAND_SIZE + 1] =
	{
		{0, 0, 0, false,},
		{0, 0, 0, false,},
		{0, 0, 0, false,},
		{0, 0, 0, false,},
		{0, 0, 0, false,}
	};

const Card_XY card_xy[HAND_SIZE] PROGMEM =
	{
		{4, 17},
		{9, 17},
		{14, 17},
		{19, 17},
		{24, 17}
	};

const char strPayout0[] PROGMEM = "HIGH CARD";
const char strPayout1[] PROGMEM = "JACKS OR BETTER";
const char strPayout2[] PROGMEM = "TWO PAIR";
const char strPayout3[] PROGMEM = "THREE OF A KIND";
const char strPayout4[] PROGMEM = "STRAIGHT";
const char strPayout5[] PROGMEM = "FLUSH";
const char strPayout6[] PROGMEM = "FULL HOUSE";
const char strPayout7[] PROGMEM = "FOUR OF A KIND";
const char strPayout8[] PROGMEM = "STRAIGHT FLUSH";
const char strPayout9[] PROGMEM = "ROYAL FLUSH";

const char* strPayoutTable[20] PROGMEM =
	{
		strPayout0, strPayout1, strPayout2, strPayout3, strPayout4, strPayout5, strPayout6, strPayout7, strPayout8, strPayout9
	};
	
const u16 valPayoutTable[10][5] PROGMEM =
	{
		{0, 0, 0, 0, 0},
		{1, 2, 3, 4, 5},
		{2, 4, 6, 8, 10},
		{3, 6, 9, 12, 15},
		{4, 8, 12, 16, 20},
		{6, 12, 18, 24, 30},
		{9, 18, 27, 36, 45},
		{25, 50, 75, 100, 125},
		{50, 100, 150, 200, 250},
		{250, 500, 750, 1000, 4000}
	};
	
const int* card_maps[DECK_SIZE + 2] PROGMEM =
	{
		spade_a, spade_2, spade_3, spade_4, spade_5, spade_6, spade_7, spade_8, spade_9, spade_t, spade_j, spade_q, spade_k,
		heart_a, heart_2, heart_3, heart_4, heart_5, heart_6, heart_7, heart_8, heart_9, heart_t, heart_j, heart_q, heart_k,
		diamond_a, diamond_2, diamond_3, diamond_4, diamond_5, diamond_6, diamond_7, diamond_8, diamond_9, diamond_t, diamond_j, diamond_q, diamond_k,
		club_a, club_2, club_3, club_4, club_5, club_6, club_7, club_8, club_9, club_t, club_j, club_q, club_k,
		card_back, card_blank
	};


/****************************************
 *			Function declarations		*
 ****************************************/
void ClearHSCursor(u8 hs_cursor_x, u8 hs_cursor_y);
void DrawBorder(u8 x, u8 y, u8 w, u8 h); 
void FadeOutVolume(void);
void SetDefaultHighScores(void);
void LoadHighScores(void);
void SaveHighScores(void);
u8 CheckHighScore(u16);
void ViewHighscoreEntry(u8 x, u8 y, u8 entry);
void FlashPressStart(void);
void SwapTitleCards(void);
void DeckInit(void);
void DeckShuffle(void);
u8 GetCard(void);
void NewHand(void);
void ReplaceCards(void);
void DrawCard(u8 x, u8 y, u8 card);
void DealCard(u8 start_x, u8 end_x, u8 y, u8 card);
void FoldCard(u8 start_x, u8 end_x, u8 y, u8 card);
void SaveVram(u8 x1, u8 y1,u8 x2, u8 y2);
void RestVram(u8 x1, u8 y1,u8 x2, u8 y2);
int GetTile(u8 xx, u8 yy);
void PrintInt2(int x,int y, unsigned int val,bool zeropad, u8 chars);
void WaitButton(void);
void PlayScreen(u8 bet);
void ShowPayoutText(u8 hand);
void ShowPayoutAmount(u8 hand, u8 bet);
void ShowPayoutTable(u8 bet);
u8 AnalyseHand(void);
void FlashWin(u8 hand, u8 bet);
void TitleScreen(void);
void ClearHelp(void);
void FancyCls(void);
void ShowStatus(u8 credits,u8 bet, u16 score);
void DealHand(void);
void ClearCards(void);

/****************************************
 *			Function definitions						*
 ****************************************/
void ClearHSCursor(u8 hs_cursor_x, u8 hs_cursor_y)
	{	
		SetTile(8 + (hs_cursor_x * 2), 7 + (hs_cursor_y * 2) - 2, 0);
		SetTile(9 + (hs_cursor_x * 2), 7 + (hs_cursor_y * 2) - 2, 0);
		SetTile(10 + (hs_cursor_x * 2), 7 + (hs_cursor_y * 2) - 2, 0);
		SetTile(8 + (hs_cursor_x * 2), 7 + (hs_cursor_y * 2) - 1, 0);
		SetTile(10 + (hs_cursor_x * 2), 7 + (hs_cursor_y * 2) - 1, 0);
		SetTile(8 + (hs_cursor_x * 2), 7 + (hs_cursor_y * 2), 0);
		SetTile(9 + (hs_cursor_x * 2), 7 + (hs_cursor_y * 2), 0);
		SetTile(10 + (hs_cursor_x * 2), 7 + (hs_cursor_y * 2), 0);
	}
	
void DrawBorder(u8 x, u8 y, u8 w, u8 h)
	{
		u8 i;
	
		SetTile(x, y, 7);
		Fill(x + 1, y, w - 2, 1, 11);
		SetTile(x + (w - 1), y, 8);
	
		for(i = 0; i < (h - 2); i++)
			{
				SetTile(x, y + 1 + i, 12);
				SetTile(x + (w - 1),y + 1 + i, 12);
			}
		
		SetTile(x, y + (h - 1), 9);
		Fill(x + 1,y + (h - 1),w - 2, 1, 11);
		SetTile(x + (w - 1), y + (h - 1), 10);
	
	}
	
void FadeOutVolume(void)
	{
		u8 i;	
		for(i=130;i;i-=2)
			{
				SetMasterVolume(i);	
				WaitVsync(1);
			}		
	}

void SetDefaultHighScores(void)
	{
		ebs.id = GAME_ID;	
		// entry #1
		ebs.data[0] = 'V';
		ebs.data[1] = 'I';
		ebs.data[2] = 'D';
		ebs.data[3] = 'E';
		ebs.data[4] = 'O';
		ebs.data[5] = ' ';
		ebs.data[6] = ' ';
		ebs.data[7] = ' ';
		ebs.data[8] = 0;
		ebs.data[9] = 150;
		// entry #2
		ebs.data[10] = 'P';
		ebs.data[11] = 'O';
		ebs.data[12] = 'K';
		ebs.data[13] = 'E';
		ebs.data[14] = 'R';
		ebs.data[15] = ' ';
		ebs.data[16] = ' ';
		ebs.data[17] = ' ';
		ebs.data[18] = 0;
		ebs.data[19] = 100;
		// entry #3
		ebs.data[20] = 'R';
		ebs.data[21] = 'E';
		ebs.data[22] = 'M';
		ebs.data[23] = 'I';
		ebs.data[24] = 'X';
		ebs.data[25] = ' ';
		ebs.data[26] = ' ';
		ebs.data[27] = ' ';
		ebs.data[28] = 0;
		ebs.data[29] = 50;
		
		EepromWriteBlock(&ebs);
	}

void LoadHighScores(void)
	{
		ebs.id = GAME_ID;

		if (EEPROM_ERROR_BLOCK_NOT_FOUND == EepromReadBlock(GAME_ID, &ebs))
			SetDefaultHighScores();
	}

void SaveHighScores(void)
	{
		ebs.id = GAME_ID;
		
		if (!isEepromFormatted())
			return;
			
		EepromWriteBlock(&ebs);	
	}

u8 CheckHighScore(u16 score)
	{
		u8 i;

		for(i = 0; i < 3; i++)
			if(score > ((ebs.data[(i * 10) + 8] * 256) + ebs.data[(i * 10) + 9]))
				return i + 1;
		return 0;
	}

void ViewHighscoreEntry(u8 x, u8 y, u8 entry)
	{
		u8 i, c;
 
		entry--;
 
		for(i=0;i<8;i++)
			{
				c = ebs.data[i + (entry * 10)];  
				PrintChar(x + i, y, c);  
			}
			
		PrintInt(x + 15, y, (ebs.data[(entry * 10) + 8] * 256) + ebs.data[(entry * 10) + 9], false);
	}

void FlashPressStart(void)
	{
		if (FlashState)
			Print(14, 26, PSTR("[?@"));
		else
			Print(14, 26, PSTR("   "));
			
		FlashState = !FlashState;
}

void SwapTitleCards(void)
	{
		if (SwapCardsState)
			{
				DrawMap(card_xy[0].x,card_xy[0].y - 14, card_v);
				DrawMap(card_xy[1].x,card_xy[1].y - 14, card_i);
				DrawMap(card_xy[2].x,card_xy[2].y - 14, card_d);
				DrawMap(card_xy[3].x,card_xy[3].y - 14, card_e);
				DrawMap(card_xy[4].x,card_xy[4].y - 14, card_o);
			}
		else
			{
				DrawMap(card_xy[0].x,card_xy[0].y - 14, card_p);
				DrawMap(card_xy[1].x,card_xy[1].y - 14, card_o2);
				DrawMap(card_xy[2].x,card_xy[2].y - 14, card_k);
				DrawMap(card_xy[3].x,card_xy[3].y - 14, card_e2);
				DrawMap(card_xy[4].x,card_xy[4].y - 14, card_r);
			}			
		SwapCardsState = !SwapCardsState;
}

										
void DeckInit(void)
	{
		u8 i;
    for (i = 0; i < DECK_SIZE; i++)
			deck[i] = i;
	}

void DeckShuffle (void)
	{
		u8 i, tmp, x;
		for (i = DECK_SIZE - 1; i > 1; i--)
			{
				x = PRNG_NEXT()%i; 
				if (x == i) continue;
				// now swap
				tmp = deck[i];
				deck[i] = deck[x];
				deck[x] = tmp;
			}
	}

u8 GetCard(void)
	{
		return deck[deck_pos++];
	}

void NewHand(void)
	{
		u8 i;
	
		for (i = 0; i < HAND_SIZE; i++)
			{	
				hand[i].card = GetCard();
				hand[i].suit = hand[i].card / 13; // Suit
				hand[i].rank = hand[i].card % 13; // Rank
				hand[i].held = false;
			}
	}

void ReplaceCards(void)
	{
		u8 i;
	
		for(i = 0; i < HAND_SIZE; i++)
				if(!hand[i].held)	
					{
						FoldCard(pgm_read_byte(&(card_xy[i].x)), DECK_POS_X, DECK_POS_Y, i);		
						hand[i].card = GetCard();
						hand[i].suit = hand[i].card / 13; // Suit
						hand[i].rank = hand[i].card % 13; // Rank
						hand[i].held = false;
					}		
		WaitVsync(20);
	}

void DrawCard(u8 x, u8 y, u8 card)
	{
		DrawMap(x, y, (int*) pgm_read_word(&(card_maps[card])));
	}

void DealCard(u8 start_x, u8 end_x, u8 y, u8 card)
	{
		u8 i;
		--end_x;
		TriggerFx(2, 130, true);
		for (i = start_x; i > end_x ; i-=2)
			{
				SaveVram(i, y, i + 3, y + 4);
				DrawCard(i, y, CARD_BACK);
				WaitVsync(1);
				RestVram(i, y, i + 3, y + 4);
			}	
		DrawCard(pgm_read_byte(&(card_xy[card].x)), pgm_read_byte(&(card_xy[card].y)), CARD_BACK);
		DrawCard(pgm_read_byte(&(card_xy[card].x)), pgm_read_byte(&(card_xy[card].y)), hand[card].card);
	
	}

void FoldCard(u8 start_x, u8 end_x, u8 y, u8 card)
	{
		u8 i;
		TriggerFx(2, 130, true);
		DrawCard(start_x, y, CARD_BLANK);
	
		for (i = start_x; i < end_x ; i+=2)
			{
				SaveVram(i, y, i + 3, y + 4);
				DrawCard(i, pgm_read_byte(&(card_xy[card].y)), hand[card].card);
				
				if (i > (DECK_POS_X - 4))
					DrawMap(DECK_POS_X, y, card_deck);
					
				WaitVsync(1);
				RestVram(i, y, i + 3, y + 4);
			}	
	}
	
void SaveVram(u8 x1, u8 y1,u8 x2, u8 y2)
	{
		u8 xi;
			for(;y1 <= y2; y1++)
				{
					for(xi = x1;xi <= x2; xi++)
						{
							vram_buff[((y2-y1) * 8) + (x2-xi)] = GetTile(xi,y1);
						}
				}	
	}

void RestVram(u8 x1, u8 y1,u8 x2, u8 y2)
	{
		u8 xi;
			for(;y1 <= y2; y1++)
				{
					for(xi = x1;xi <= x2; xi++)
						{
							SetTile(xi,y1,vram_buff[((y2-y1) * 8) + (x2-xi)]);	   
						}
				}
	}

int GetTile(u8 xx, u8 yy)
	{
		int i1 = (yy * 80) + (xx * 2);
		return(((vram[i1 + 1] * 256) + vram[i1] - (int)(tiles)) / 48);
	}

void WaitButton(void)
	{
		while (ReadJoypad(0) != BTN_START)
		{
		};
	}

void PrintInt2(int x,int y, unsigned int val,bool zeropad, u8 chars)
	{
		unsigned char c,i;

		for(i=0;i<chars;i++)
			{
				c=val%10;
				if(val>0 || i==0)
					{
						SetFont(x--,y,c+CHAR_ZERO);
					}
				else
					{
						if(zeropad)
							{
								SetFont(x--,y,CHAR_ZERO);
							}
						else
							{
								SetFont(x--,y, 0);
							}
					}
		val=val/10;
			}		
	}

void PlayScreen(u8 bet_amount)
		{
			u8 i;
						
			ClearVram();
			DrawMap(0, 0, map_play);
			
			Print(7, 1, PSTR("UZEBOX VIDEO POKER REMIXED"));
			Print(7, 3, PSTR("HAND"));
			Print(19, 3, PSTR("WIN"));
			Print(29, 3, PSTR("HELP"));
			
			for (i = 0; i < MAX_PAYOUT; i++)
				{
					ShowPayoutText(i);
					ShowPayoutAmount(i, bet_amount);
				}
				
			Print(CREDIT_TEXT_X, CREDIT_TEXT_Y, PSTR("CREDIT"));
			Print(BET_TEXT_X, BET_TEXT_Y, PSTR("BET "));
			Print(WIN_TEXT_X, WIN_TEXT_Y, PSTR("WIN"));			
		}
		
void ShowPayoutText(u8 hand)
	{
		Print(PAYOUT_TEXT_X, PAYOUT_TEXT_Y - hand, (char*) pgm_read_word(&(strPayoutTable[hand + 1])));
	}
	
void ShowPayoutAmount(u8 hand, u8 bet)
	{
		PrintInt2(PAYOUT_AMOUNT_X, PAYOUT_AMOUNT_Y - hand, pgm_read_word(&(valPayoutTable[hand + 1][bet])), false, 4);
	}
	
void ShowPayoutTable(u8 bet)
	{
		u8 i;
		
		for (i = 0; i < MAX_PAYOUT; i++)
			ShowPayoutAmount(i, bet);
	}
	
u8 AnalyseHand(void)
	{
		u8 i,j, tmp;
		u8 card_count[13];
		bool jacks_or_better = false;
		bool flush = true;
		bool low_straight = true;
		bool high_straight = false;
		
		//Quick Sort card ranks
		for(i = 0; i < 5; i++)
			{
				for(j = (i + 1); j < 5; j++)
					{
						if(hand[i].rank > hand[j].rank)
							{
								hand[HAND_SIZE].card = hand[i].card;
								hand[HAND_SIZE].suit = hand[i].suit;
								hand[HAND_SIZE].rank = hand[i].rank;
								hand[HAND_SIZE].held = hand[i].held;
								
								hand[i].card = hand[j].card;
								hand[i].suit = hand[j].suit;
								hand[i].rank = hand[j].rank;
								hand[i].held = hand[j].held;
								
								hand[j].card = hand[HAND_SIZE].card;
								hand[j].suit = hand[HAND_SIZE].suit;
								hand[j].rank = hand[HAND_SIZE].rank;
								hand[j].held = hand[HAND_SIZE].held;								
							}
					}
			}
		
		// Init card counts
		for (i = 0; i < 13; i++)
			card_count[i] = 0;
		
		// Count number of each face card
		for (i = 0; i < HAND_SIZE; i++)
			card_count[hand[i].rank]++;
		
		// Test for pair of Jacks or better
		if ((card_count[ACE] > 1) || (card_count[JACK] > 1) || (card_count[QUEEN] > 1) || (card_count[KING] > 1))
			jacks_or_better = true;
		
		//Quick Sort card counts
		for(i = 0; i < 13; i++)
			{
				for(j = (i + 1); j < 13; j++)
					{
						if(card_count[i] > card_count[j])
							{
								tmp = card_count[i];
								card_count[i] = card_count[j];
								card_count[j] = tmp;
							}
					}
			}
		
		if (card_count[12] == 4)
			return FOUR_OF_A_KIND;
		else if ((card_count[12] == 3) && (card_count[11] == 2))
			return FULL_HOUSE;
		else if ((card_count[12] == 3) && (card_count[11] == 1))
			return THREE_OF_A_KIND;
		else if ((card_count[12] == 2) && (card_count[11] == 2))
			return TWO_PAIR;
		else if ((card_count[12] == 2) && jacks_or_better)
			return JACKS_OR_BETTER;
	
		// Test for a flush
		for(i = 1; i < 5; i++)
			if(hand[0].suit != hand[i].suit)
				flush = false;
	
		//Test for a straight <> 10,J,Q,K,A
		for(i = 0; i < 4; i++)
			if(hand[i].rank != hand[i + 1].rank - 1)
				low_straight = false;
	
		if(flush && low_straight)
			return STRAIGHT_FLUSH;

		//Test for a straight = 10,J,Q,K,A
		if((hand[0].rank == ACE) && (hand[1].rank == TEN) && (hand[2].rank == JACK) && (hand[3].rank == QUEEN) && (hand[4].rank == KING))
			high_straight = true;
	
		if(flush && high_straight)
			return ROYAL_FLUSH;
	
		if(flush)
			return FLUSH;
		
		if(low_straight || high_straight)
			return STRAIGHT;
		
		// No winning hand found
		return 0;
	}

void FlashWin(u8 hand, u8 bet)
	{
		u8 i;
		
		for (i = 0; i < 6; i++)
			{
				Fill(PAYOUT_TEXT_X, PAYOUT_TEXT_Y - hand, 15, 1, 0);
				Fill(PAYOUT_AMOUNT_X - 3, PAYOUT_AMOUNT_Y - hand, 4, 1 , 0);
				WaitVsync(10);
				ShowPayoutText(hand);
				ShowPayoutAmount(hand, bet);
				WaitVsync(20);
			}
	}

void TitleScreen(void)
	{
		u8 i;
		
		DrawMap(1, 1, map_table);
		DrawBorder(10, 15, 20, 9);
		
		Print(5, 10, PSTR("REMIXED BY JIM CROSKELL MMXIII"));
		Print(14, 13, PSTR("HALL OF FAME"));
		Print(8, 26, PSTR("PRESS [?@ BUTTON TO PLAY"));

		SetMasterVolume(130);
		StartSong(midisong);
		LoadHighScores();
	
		for(i =0 ; i < 3; i++)
			ViewHighscoreEntry(12, (i * 2) + 17, i + 1);
	}

void ClearHelp(void)
	{
		Fill(HELP_X_POS, HELP_Y_POS, HELP_X_SIZE, HELP_Y_SIZE, 0);
	}

void FancyCls(void)
	{
		u8 i;
		
		for (i = 0; i < 20; i++)
			{
				Fill(i, 0, 1, 28, 0);
				WaitVsync(1);
				Fill(39 - i, 0, 1, 28, 0);
				WaitVsync(1);
			}
	}

void ShowStatus(u8 credits,u8 bet, u16 score)
	{
		PrintByte(CREDIT_VAL_X, CREDIT_VAL_Y, credits, true);
		PrintChar(BET_VAL_X, BET_VAL_Y, 48 + bet);
		PrintInt(WIN_VAL_X, WIN_VAL_Y, score, true);
	}
	
void DealHand(void)
	{
		u8 i;
		
		for (i = 0; i <  HAND_SIZE; i++)
			if (!hand[i].held)
				DealCard(DECK_POS_X, pgm_read_byte(&(card_xy[i].x)), DECK_POS_Y, i);
	}

void ClearCards(void)
	{
		u8 i;
	
		for(i = 0; i < HAND_SIZE; i++)
			{
					FoldCard(pgm_read_byte(&(card_xy[i].x)), DECK_POS_X, DECK_POS_Y, i);		
			}	
	}
	
/************************
 *			Main 		*
 ************************/
int main(void)
	{
		u16 btnPrev = 0;
		u16 btnHeld = 0;
		u16 btnPressed = 0;
		u16 btnReleased = 0;
		u8 i, tmp;
		u8 bet = 0;
		u8 credits = 0;
		u16 score = 0;
		u8 current_card = 0;
		u8 flashStartTimer = 0;
		u8 swapCardsTimer = 0;
		u8 high_score_pos = 0;
		u8 hs_cursor_x;
		u8 hs_cursor_y;
		u8 hs_name_len;
		char hs_name[8];
			
		// Initialise
		SetFontTable(font);
		SetTileTable(tiles);
		ClearVram();
		SetMasterVolume(130);
		InitMusicPlayer(patches);
		SetMasterVolume(130);
		LoadHighScores();	
		gameState = titleDraw;
		flashStartTimer = 1;
		swapCardsTimer = 1;
		DeckInit();
		
		while(true)
			{
				if (GetVsyncFlag())
					{
						ClearVsyncFlag();
			
						btnHeld = ReadJoypad(0);
						btnPressed = btnHeld&(btnHeld^btnPrev);
						btnReleased = btnPrev&(btnHeld^btnPrev);
						btnPrev = btnHeld;

						switch (gameState)
							{
								case titleDraw:
									TitleScreen();
									gameState = titleWait;
								break;
								
								case titleWait:						
									if (btnHeld & BTN_START)
										{
											Print(14, 26, PSTR("[?@"));
											FancyCls();
											FadeOutVolume();	
											StopSong();
											gameState = playStart;
										}
									else if (btnHeld ==  BTN_SL + BTN_SELECT + BTN_X)
										{
											SetDefaultHighScores();
											for(i =0 ; i < 3; i++)
												ViewHighscoreEntry(12, (i * 2) + 17, i + 1);
										}
									else
										{
											if (--flashStartTimer == 0)
												{
													FlashPressStart();
													flashStartTimer = HZ>>2;
												}

											if (--swapCardsTimer == 0)
												{
													SwapTitleCards();
													swapCardsTimer = HZ;
												}		
										}
										
									++prng;
									prng = MAX(prng,1);
								break;

								case playStart:
									ClearVram();
									bet = BET_MIN;
									credits = CREDITS_START;
									score = SCORE_START;
									PlayScreen(bet);
									gameState = playing;
								break;
								
								case playing:
									SetMasterVolume(130);	
									bet = MIN(bet, credits);
									ClearHelp();
									Print(HELP_X_POS + 2, HELP_Y_POS + 3, PSTR("<>=>   BET"));
									Print(HELP_X_POS + 2 , HELP_Y_POS + 5, PSTR("[?@   DEAL"));
									PRNG_NEXT();
									DeckShuffle();
									deck_pos = 0;
									NewHand();									
									do
										{
											ShowStatus(credits - bet, bet, score);
											ShowPayoutTable(bet - 1);
											
											btnHeld = ReadJoypad(0);
											btnPressed = btnHeld & (btnHeld^btnPrev);
											btnReleased = btnPrev & (btnHeld^btnPrev);
											btnPrev = btnHeld;
							
											switch (btnPressed)
												{
													case BTN_SL:
														TriggerFx(6, 130, false);
														--bet;
														bet=MAX(bet, BET_MIN);
													break;
													
													case BTN_SR:
														TriggerFx(6, 130, false);
														if((credits - bet) > 0)
															{
																++bet;
																bet=MIN(bet, BET_MAX);
															}
													break;
												}
										}
									while(btnPressed != BTN_START);
									
									ClearHelp();
									DealHand();
									credits -= bet;
									Print(HELP_X_POS + 2, HELP_Y_POS + 2, PSTR("*+  SELECT"));
									Print(HELP_X_POS + 2 , HELP_Y_POS + 4, PSTR(",     HOLD"));
									Print(HELP_X_POS + 2 , HELP_Y_POS + 6, PSTR("[?@   CONT"));
									current_card = 3;
									
									do
										{
											DrawMap(CARD_CURSOR_X_POS + ((current_card -1) * 5), CARD_CURSOR_Y_POS, map_card_cursor);
											
											btnHeld = ReadJoypad(0);
											btnPressed = btnHeld&(btnHeld^btnPrev);
											btnReleased = btnPrev&(btnHeld^btnPrev);
											btnPrev = btnHeld;
							
											switch (btnPressed)
												{
													case BTN_LEFT:
														TriggerFx(6, 130, false);
														Fill(CARD_CURSOR_X_POS + ((current_card -1) * 5), CARD_CURSOR_Y_POS, 2, 1, 0x9a);
														--current_card;
														current_card = MAX(current_card, 1);
													break;
													
													case BTN_RIGHT:
														TriggerFx(6, 130, false);
														Fill(CARD_CURSOR_X_POS + ((current_card -1) * 5), CARD_CURSOR_Y_POS, 2, 1, 0x9a);
														++current_card;
														current_card = MIN(current_card, 5);
													break;
													
													case BTN_A:
														if(hand[current_card - 1].held)
															{
																TriggerFx(8, 130, false);
																Fill(CARD_CURSOR_X_POS + ((current_card -1) * 5 - 1), CARD_CURSOR_Y_POS - 6, 4, 1, 0x9a);
															}
														else
															{
																TriggerFx(7, 130, false);
																DrawMap(CARD_CURSOR_X_POS + ((current_card -1) * 5 - 1), CARD_CURSOR_Y_POS - 6, map_card_held);
															}
					
														hand[current_card - 1].held = !hand[current_card - 1].held;
													break;
												}							
										}
									while(btnPressed != BTN_START);
						
									Fill(CARD_CURSOR_X_POS + ((current_card -1) * 5), CARD_CURSOR_Y_POS, 2, 1, 0x9a);

									ClearHelp();
									ReplaceCards();
									DealHand();
									
									tmp = AnalyseHand();

									if (tmp)
										{
											Print(HELP_X_POS + 4, HELP_Y_POS + 2, PSTR("YOU WIN"));
											Print(HELP_X_POS + 3, HELP_Y_POS + 4, PSTR("NICE HAND"));
											Print(HELP_X_POS + 3, HELP_Y_POS + 6, PSTR("WELL DONE"));
											TriggerFx(4, 100 ,true);
											FlashWin(tmp - 1, bet - 1);
											score += pgm_read_word(&(valPayoutTable[tmp][bet - 1]));
										}
									else
										{
											Print(HELP_X_POS + 3, HELP_Y_POS + 2, PSTR("YOU LOSE"));
											Print(HELP_X_POS + 2, HELP_Y_POS + 4, PSTR("BETTER LUCK"));
											Print(HELP_X_POS + 3, HELP_Y_POS + 6, PSTR("NEXT HAND"));
											TriggerFx(5, 100, true);
											WaitVsync(120);
										}
									Fill(CARD_CURSOR_X_POS - 1, CARD_CURSOR_Y_POS - 6, 24, 1, 0x9a);	
									ClearCards();
									
									if (credits == 0)
										{
											ClearHelp();
											ShowStatus(0, bet, score);
											Print(HELP_X_POS + 2, HELP_Y_POS + 3, PSTR("GAME  OVER"));
											Print(HELP_X_POS + 2, HELP_Y_POS + 5, PSTR("[?@   CONT"));
											
											WaitButton();
						
											//high_score_pos = 0;
											high_score_pos = CheckHighScore(score);
											
											if(high_score_pos)
												gameState = highScore;
											else
												gameState = gameOver;
										}		
								break;
									
								case highScore:
									FancyCls();
									SetMasterVolume(130);
									StartSong(hi_score);
					
									// Move old scores down if necessary
									if(high_score_pos <3)
										{
											for(i = 0; i < 10; i++)
												ebs.data[20 + i] = ebs.data[10 + i];
						
											if(high_score_pos == 1)
												for(i = 0; i < 10; i++)
													ebs.data[10 + i] = ebs.data[i];
										}
						
									DrawBorder(9, 6, 21, 9);
									DrawBorder(13, 18, 13, 3);					
									Print(12, 2, PSTR("NEW HIGH SCORE&"));
									PrintInt(21, 4, score, true);
									Print(11, 8, PSTR("A B C D E F G H I"));
									Print(11, 10, PSTR("J K L M N O P Q R"));
									Print(11, 12, PSTR("S T U V W X Y Z  "));
									Print(12, 16, PSTR("ENTER YOUR NAME"));
									Print(11, 23, PSTR("!\"*+    MOVE CURSOR"));
									Print(11, 24, PSTR(",       SELECT"));
									Print(11, 25, PSTR(";       BACKSPACE"));
									Print(11, 26, PSTR("[?@     DONE"));
									
									hs_cursor_x = 1;
									hs_cursor_y = 1;
									hs_name_len = 0;
									for(i = 0; i < 8; i++)
										hs_name[i] = ' ';
									do
										{
											// Draw cursor
											SetTile(8 + (hs_cursor_x * 2), 7 + (hs_cursor_y * 2) - 2, 13);
											SetTile(9 + (hs_cursor_x * 2), 7 + (hs_cursor_y * 2) - 2, 14);
											SetTile(10 + (hs_cursor_x * 2), 7 + (hs_cursor_y * 2) - 2, 15);
											SetTile(8 + (hs_cursor_x * 2), 7 + (hs_cursor_y * 2) - 1, 16);
											SetTile(10 + (hs_cursor_x * 2), 7 + (hs_cursor_y * 2) - 1, 16);
											SetTile(8 + (hs_cursor_x * 2), 7 + (hs_cursor_y * 2), 17);
											SetTile(9 + (hs_cursor_x * 2), 7 + (hs_cursor_y * 2), 14);
											SetTile(10 + (hs_cursor_x * 2), 7 + (hs_cursor_y * 2), 18);
											
											for(i = 0; i < 8; i++)
												PrintChar(15 + i, 19, hs_name[i]);
											
											Print(15 + hs_name_len, 19, PSTR("% "));
											WaitVsync(1);
	
											btnHeld = ReadJoypad(0);
											btnPressed = btnHeld & (btnHeld^btnPrev);
											btnReleased = btnPrev & (btnHeld^btnPrev);
											btnPrev = btnHeld;
							
											switch (btnPressed)
												{
													case BTN_UP:
														TriggerFx(6, 130, false);
														ClearHSCursor(hs_cursor_x, hs_cursor_y);
														--hs_cursor_y;
														hs_cursor_y = MAX(hs_cursor_y, 1);							
													break;
													
													case BTN_DOWN:
														TriggerFx(6, 130, false);
														ClearHSCursor(hs_cursor_x, hs_cursor_y);
														++hs_cursor_y;
														hs_cursor_y = MIN(hs_cursor_y, 3);
													break;
													
													case BTN_LEFT:
														TriggerFx(6, 130, false);
														ClearHSCursor(hs_cursor_x, hs_cursor_y);
														--hs_cursor_x;
														hs_cursor_x = MAX(hs_cursor_x, 1);														
													break;
													
													case BTN_RIGHT:
														TriggerFx(6, 130, false);
														ClearHSCursor(hs_cursor_x, hs_cursor_y);
														++hs_cursor_x;
														hs_cursor_x = MIN(hs_cursor_x, 9);														
													break;
													
													case BTN_A:
														if (hs_name_len < 8)
															{
																TriggerFx(7, 130, false);
																hs_name[hs_name_len] = 65 + (((hs_cursor_y - 1) * 9) + (hs_cursor_x - 1));
														
																if(hs_name[hs_name_len] == 91)
																hs_name[hs_name_len] = ' ';
															
																++hs_name_len;
															}
														
														hs_name_len = MIN(hs_name_len, 8);
													break;
													
													case BTN_X:
														if(hs_name_len > 0)
															{
																hs_name[hs_name_len - 1] = ' ';
																TriggerFx(8, 130, false);
																--hs_name_len;
															}
													break;
												}
										}
									while(btnPressed != BTN_START);
					
									// Write new name
									for(i = 0; i < 8; i++)
										ebs.data[((high_score_pos - 1) * 10) + i] = hs_name[i];
						
									// Write high score & save to eeprom
									ebs.data[((high_score_pos - 1) * 10) + 8] = score / 256;
									ebs.data[((high_score_pos - 1) * 10) + 9] = score % 256;
									SaveHighScores();

									gameState = gameOver;
								break;
								
								case gameOver:
									FancyCls();
									flashStartTimer = 1;
									gameState = titleDraw;
								break;
							}		
					}
			}	
			
	}
